#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "libdtm.h"
 
/* Constants for the LLA / UTM conversions (variables names are germanish --
   these functions have been provided by TUB during the Comets projet) */
#define a_klein	     (double)(6378137.0) /* Aequatorialradius in [m], = grosse Halbachse */   
#define f	     (double)(0.0033528107) /* Abplattung */
#define ko	     (double)(0.9996)

/* 
 * These functions provide ways to set a geodetic reference to a DTM. The
 * following abbreviations and conventions stand:
 * LLA: latitude, longitude, altitude
 * UTM: Universal tranverse mercator
 * The X axis of the DTM correspond to the line coordinates(*) and so the Y axis
 * correspond to the columns (lines and columns are respectively indexed by the
 * variables i and j in most of the library functions).
 * X points eastwards, Y points northwards, Z grows vertically (ENU convention)
 * 
 * All the geodetic coordinate transform assume a WGS84 geoid.
 * 
 * (*) (the 'stockage' field of the DTM is _systematically_ 'X_LIGNES' in all
 * the library functions, although the symbol 'X_COLONNES' has been defined),
 */

/**********************************************************************
 *
 * Geo-reference a DTM with LLA coordinates of its top-left cell (that is, south
 * west cell, or (0,0) coordinates).
 *
 * lat and lon parameters must be given in decimal degrees
 * 
 **********************************************************************/

int 
dtm_setLLAGeodesicReference (DTM*                dtm,
			     int                 consideredPoint,
			     double              lat,
			     double              lon,
			     double              alt)
{
  int i_nextEastMeridian;	/* In degrees : -9, -3, 3, 9, 15, ...*/

  if (consideredPoint != 0)	/* The geodetic reference must correspond to
				   the origin of the DTM, not to the cell with
				   cooridnates (0,0) */
    return dtmError(ERR_DTM_FCT_NON_IMPLEMENTEE);
  else { 
    dtm->geodesicOrigin.lat = DEG_TO_RAD(lat);
    dtm->geodesicOrigin.lon = DEG_TO_RAD(lon);
    dtm->geodesicOrigin.alt = alt;
    dtm->geoOrientation = GEO_EAST;
    dtm->geoReference = GEO_REFERENCE_SET;
    
    /* Find the nextEastMeridian : a multiple of 6 degrees, starting
       at 3 degrees (e.g. 3, 9, 15 degrees) in the positive range (at
       -3 degrees in the negative range) */
    i_nextEastMeridian = 6 * (floor ((lon + 3.0)/6.0));
    dtm->geodesicOrigin.nextEastMeridian = 
      DEG_TO_RAD(i_nextEastMeridian + 3.0);
    
    /* Computes the corresponding UTM coordinates */
    dtm_wgs84lla2utm(dtm->geodesicOrigin.lon, 
		     dtm->geodesicOrigin.lat,
		     dtm->geodesicOrigin.nextEastMeridian,
		     &(dtm->geodesicOrigin.easting),
		     &(dtm->geodesicOrigin.northing));
  }

  return (DTM_OK);
}

/**********************************************************************
 *
 * Geo-reference a DTM with UTM coordinates. The coordinates can correspond
 * either to the origin of the DTM or to its top-left cell (i.e. the one with
 * coordinates (0,0) dem cells array), depending on the value of the
 * 'consideredPoint' variable.
 *
 **********************************************************************/

int 
dtm_setUTMGeodesicReference (DTM*                dtm,
			     int                 consideredPoint,
			     double              easting,
			     double              northing,
			     double              alt,
			     double              nextEast)
{
  if (consideredPoint != 0) {
    /* The geodetic coordinates are provided for the origin of the DTM */ 
    dtm->geodesicOrigin.easting = easting + dtm->x_origine;
    dtm->geodesicOrigin.northing = northing + dtm->y_origine;
  } else {
    /* The geodetic coordinates are the ones of the "top-left" pixel of the DTM,
       i.e. indices (0,0) */ 
    dtm->geodesicOrigin.easting = easting;
    dtm->geodesicOrigin.northing = northing;
  }
  dtm->geodesicOrigin.alt = alt;
  dtm->geoOrientation = GEO_EAST;
  dtm->geoReference = GEO_REFERENCE_SET;
  dtm->geodesicOrigin.nextEastMeridian = DEG_TO_RAD(nextEast);
  
  /* Computes the corresponding LLA coordinates */
  dtm_wgs84utm2lla (dtm->geodesicOrigin.easting,
		    dtm->geodesicOrigin.northing, 
		    dtm->geodesicOrigin.nextEastMeridian,
		    &(dtm->geodesicOrigin.lat), 
		    &(dtm->geodesicOrigin.lon));

  return (DTM_OK);
}

/**********************************************************************
 *
 *  LLA <-> UTM conversions functions (provided by TUB during the Comets
 *  project) 
 *
 **********************************************************************/

/* parameters: lon_rad, lat_rad, lon0_rad in radians with lon0 giving
   the next eastern meridian (3,9,15...). Sets the UTM easting and
   northing */ 
void
dtm_wgs84lla2utm(double lon_rad, 
		 double lat_rad, 
		 double lon0_rad,
		 double* easting,
		 double* northing) 
{
  double b, e, e_strich_2, N, T, a_gross, C, M;
  double e2,e4,e6;
  double a_gross2, a_gross3, a_gross4, a_gross5, a_gross6;

  b             = (double)(a_klein * (1.0 - f));
  e             = sqrt(1.0 - (b*b) / (a_klein*a_klein));
  e2 = e*e;
  e4 = e2*e2;
  e6 = e4*e2;
  e_strich_2	= (double)(e2 / (1.0 - e2));
  N		= (double)(a_klein / sqrt( 1.0 - e2 * ( sin(lat_rad)*sin(lat_rad) ) ) );
  T		= (double)((tan(lat_rad)) * (tan(lat_rad)));
  a_gross	= (double)((lon_rad - lon0_rad) * cos(lat_rad));
  a_gross2 = a_gross  * a_gross;
  a_gross3 = a_gross2 * a_gross;
  a_gross4 = a_gross3 * a_gross;
  a_gross5 = a_gross4 * a_gross;
  a_gross6 = a_gross5 * a_gross;
  C		= (double)(e_strich_2 * (cos(lat_rad)*cos(lat_rad)));
  M 		= a_klein * ((1.0 - e2 / 4.0 - 3.0 * e4 / 64.0 - 5.0 * e6 / 256.0) * lat_rad - 
			     (3.0 * e2 / 8.0 + 3.0 * e4 / 32.0 + 45.0 * e6 / 1024.0) * sin(2.0 * lat_rad) + 
			     (15.0 * e4 / 256.0 + 45.0 * e6 / 1024.0) * sin(4.0 * lat_rad) - 
			     (35.0 * e6 / 3072.0) * sin(6.0 * lat_rad));

  *easting = (double)500000.0 + ko * N * (a_gross + (1.0 - T + C) * a_gross3 / 6.0 + 
				      (5.0 - 18.0 * T + T*T + 72.0 * C - 58 * e_strich_2) * a_gross5 / 120.0);
  *northing = (double)ko * (M + N * tan(lat_rad) * (a_gross2 / 2.0 + (5.0 - T + 9.0 * C + 4.0 * (C*C)) * 
						a_gross4 / 24.0 + (61.0 - 58.0 * T + T*T + 600.0 * C - 330.0 * e_strich_2) * a_gross6 / 720.0));

  return;
}

/* parameters: x, y in meters for UTM easting and northing (both false)
meridian as described abov, lat, lon in radians */
void 
dtm_wgs84utm2lla (double x, 
		  double y, 
		  double meridian,
		  double* lat,
		  double* lon)
{
  double b, e, e_strich_2, phi0, lam0, M0, M, mue, e1, phi1, C1, N1, T1, R1, D, phi, lam;
  double e2, e4, e6;
  double e12, e13, e14;
  double D2,D3,D4,D5,D6;

/*    false easting */
  x -= 500000.0;

  b             = (double)(a_klein * (1.0 - f));
  e             = (double)(sqrt(1.0 - (b*b) / (a_klein*a_klein)));
  e2 = e*e;
  e4 = e2*e2;
  e6 = e2*e4;
  e_strich_2	= (double)(e2 / (1.0 - e2));

  phi0		= 0.0;
  lam0		= meridian;

  M0		= a_klein * ((1.0 - e2/4.0 - 3.0*e4/64.0 - 5.0*e6/256.0) * phi0 - 
				     (3.0*e2/8.0 + 3.0*e4/32.0 + 45.0*e6/1024.0) * sin(2.0*phi0) + 
				     (15.0*e4/256.0 + 45.0*e6/1024) * sin(4.0*phi0) + 
				     (35.0*e6/3072.0) * sin(6.0*phi0));

  M 		= M0 + y/ko;
  mue 		= M/(a_klein * (1.0 - e2/4.0 - 3.0*e4/64.0 - 5.0*e6/256.0));
  
  e1		= (1.0-sqrt(1.0-e2))/(1.0+sqrt(1.0-e2));
  e12 = e1*e1;
  e13 = e12*e1;
  e14 = e13*e1;

  phi1 		= mue + (3.0*e1/2.0 - 27.0*e13/32.) * sin(2.0*mue) + 
    (21.0*e12/16.0 - 55.0*e14/32.0) * sin(4.0*mue) + 
    (151.0*e13/96.) * sin(6.0*mue) + 
    (1097.0*e14/512.0) * sin(8.0*mue);

  C1		= (double)(e_strich_2 * (cos(phi1)*cos(phi1)));
  N1		= (double)(a_klein / sqrt(1.0 - e2 * ( sin(phi1)*sin(phi1) ) ) );
  T1		= (double)((tan(phi1))*(tan(phi1)));
  R1		= a_klein * (1.0-e2)/ pow ( 1.0-e2* ( sin(phi1)*sin(phi1) ) , 1.5);
  D		= x/(N1*ko);
  D2 = D*D;
  D3 = D2*D;
  D4 = D3*D;
  D5 = D4*D;
  D6 = D5*D;

  phi 		= phi1 - (N1*tan(phi1)/R1) * (D2/2.0 - (5.0 + 3.0 * T1 + 10.0 * C1 - 4.0 * (C1*C1) - 9.0 * e_strich_2) * D4/24.0 + 
					      (61.0 + 90.0 * T1 + 298.0 * C1 + 45.0 * (T1*T1) - 252 * e_strich_2 - 3.0 * (C1*C1)) * D6/720.0);

  lam			= lam0 + (D - (1.0 + 2.0 * T1 + C1) * D3/6.0 + 
				  (5.0 - 2.0*C1 + 28.0 * T1 - 3.0*(C1*C1)+8.0*e_strich_2+24.0*(T1*T1)) * D5/120.0)/cos(phi1);

  *lat = phi;
  *lon = lam;

  return;
}


/**********************************************************************
 *
 * LLA -> LambertIIe conversion, kindly provided by Roland.Lenain@cemagref.fr
 * (hence comments are in french)
 *
 * Warning: latitude and longitude must be provided in decimal degrees.
 *
 **********************************************************************/

int
dtm_calculZoneLambert_ (double latitudeWGS84, 
			double longitudeWGS84)
{
  double SeuilsZonesLambert_[5]= {51.3,48.15,45.45,42.3,0.0};
  int zoneLambert=0;

  while( zoneLambert<5 && latitudeWGS84 < SeuilsZonesLambert_[zoneLambert] ) 
    ++zoneLambert;
  if(zoneLambert==5) zoneLambert=0;
  
  return zoneLambert;
}

void
dtm_lla2LambertIIe (double latitudeWGS84,
		    double longitudeWGS84,
		    double* latitudeLambert, 
		    double* longitudeLambert)
{
  double CA, CB, CE, DECAL;
  double lat, lon;
  double lam1, phi1, N1, ro1;
  double phio, KO, CX, CY, Loo, gamma, L, Ro, R;
  int zoneLambert_;
  
  zoneLambert_ = dtm_calculZoneLambert_(latitudeWGS84, longitudeWGS84);

  // Paramtres de l'ellipsoide WGS84
  const double a1 = 6378137.0;  
  const double f1 = 0.00335281066;       

  // Paramtres de l'ellipsoide NTF (Clarke 1880)
  const double a2 = 6378249.2; 
  const double f2 = 0.00340754952;
  const double e2 = 0.00680348764;

  const int tx   = 168;
  const int ty   = 60;
  const int tz   = -320;

  lat = latitudeWGS84*M_PI/180;
  lon = longitudeWGS84*M_PI/180;

  // WGS84->NTF
  lam1 = lon;
  phi1 = lat;
  N1   = a1 / ( sqrt(1 - e2 * sin(phi1) * sin(phi1) ) );
  ro1  = a1 * (1-e2) / pow(1-e2 * sin(phi1),1.5);
  lon += ( -tx*sin(lam1) + ty*cos(lam1) ) / ( N1*cos(phi1) );
  lat += ( -tx*sin(phi1)*cos(lam1) 
      -ty*sin(phi1)*sin(lam1)
      +tz*cos(phi1) + ( a2*(f2-f1) + f2*(a2-a1) ) * sin(2*phi1) ) / ro1;

  CA    = 6378249.2;
  CB    = 6356515.0;
  CE    = sqrt( ((CA*CA)-(CB*CB)) / (CA*CA));  
  DECAL = 0.02596921296 * M_PI_2;

  switch(zoneLambert_)
  {
    case 1:
      phio = 0.55 * M_PI_2;
      KO   = 0.999877340;
      CX   = 600000;
      CY   = 200000;
      break;
    case 2:
      phio = 0.52 * M_PI_2;
      KO   = 0.999877420;
      CX   = 600000;
      CY   = 200000;
      break;
    case 3:
      phio = 0.49 * M_PI_2;
      KO   = 0.999877500;
      CX   = 600000;
      CY   = 200000;
      break;
    case 4:
      phio = 0.4685 * M_PI_2;
      KO   = 0.999944710;
      CX   = 234.358;
      CY   = 4185861.369;
      break;
    default: phio=0;
      KO   = 0;
      CX   = 0;
      CY   = 0;
  }
  Loo   = log(tan(phio/2+M_PI_4))+CE/2*log((1-CE*sin(phio))/(1+CE*sin(phio)));
  gamma = (lon-DECAL)*sin(phio);
  L     = log(tan(lat/2+M_PI_4))+CE/2*log((1-CE*sin(lat))/(1+CE*sin(lat)));
  Ro    = (CA*KO) / ( sqrt(1-CE*CE*sin(phio)*sin(phio))*tan(phio));
  R     = Ro * exp(-sin(phio)*(L-Loo));

  *latitudeLambert  = Ro-R*cos(gamma)+CY;
  *longitudeLambert = R*sin(gamma)+CX;
}

