/* Les includes génériques de C */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* Les structures, protos et erreurs de la lib */
#include "libdtm.h"

/**********************************************************************
 *
 * Affichage à l'écran d'infos sur une structure "DTM"
 *
 **********************************************************************/

int
info_dtm (DTM*	mnt)

{
  /* Infos sur la géométrie du MNT */
  printf ("\n\t\tThe DTM has %d lines and %d columns\n",
	  mnt->nblig, mnt->nbcol);
  
  printf ("\tOccupied memory: %d Octets\n",
	  (int) (mnt->nblig * mnt->nbcol * sizeof (DTM_CELL)));
  printf ("\t(or %.1f Ko or %.1f Mo)\n",
	  ((double) mnt->nblig * mnt->nbcol * sizeof (DTM_CELL)) / 1024.0,
	  ((double) mnt->nblig * mnt->nbcol * sizeof (DTM_CELL)) 
	  / (1024.0 * 1024.0));
  
  printf ("\tSizes along the X and Y axes: (%.2f, %.2f)\n", 
	  mnt->size_x, mnt->size_y);
  printf ("\tScales along the X and Y axes: (%.3f, %.3f)\n", 
	  mnt->scale_x, mnt->scale_y);
  printf ("\tCoordinates of the 'top-left' point of the first DTM cell\n"
	  "\tin the global DTM frame: (%.3f %.3f)\n",
	  mnt->x_origine, mnt->y_origine);
  
  switch (mnt->type) {
  case DTM_FIXED:
    printf ("\tThe DTM is tied to the world\n"); 
    break;
  case DTM_SLIDING:
    printf ("\tThe DTM slides along the robot\n"); 
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }

  switch (mnt->stockage) {
  case X_LIGNES:
    printf ("\tThe X axis is associated to the lines (i indices)\n"); 
    break;
  case X_COLONNES:
    printf ("\tThe X axis is associated to the columns (j indices)\n"); 
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }
  
  printf ("\tThe origin point of the world frame belongs\n"
	  "\tto the cell: (%d, %d)\n",
	  mnt->i_origine, mnt->j_origine);
    
  /* Référence géodésique */
  switch (mnt->geoReference) {
  case GEO_REFERENCE_SET:
    printf ("\tGeodesic reference of the origin - top-left corner of the cell (0,0):\n");
    dtm_infoGeodesicPos (&(mnt->geodesicOrigin));

    printf ("\t\tThe X axis is oriented towards ");
    switch (mnt->geoOrientation) {
    case NO_GEO_ORIENTATION:
      printf ("Somewhere\n");
      break;
    case GEO_NORTH:
      printf ("North\n");
      break;
    case GEO_WEST:
      printf ("West\n");
      break;
    case GEO_SOUTH:
      printf ("South\n");
      break;
    case GEO_EAST:
      printf ("East\n");
      break;
    default:
      printf ("Nowhere !\n");
      return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
      /*     break; */
      /* Unreached */
    }
    break;
  case NO_GEO_REFERENCE:
    printf ("\tNo geodesic reference for the origin\n");
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }

  /* Des images 3d ont été fusionnées ? */
  if (mnt->nb_im3d)
    printf ("\t%d 3D images have been merged in the DTM\n", mnt->nb_im3d);
  else
    printf ("\t(no 3D image have been merged in the DTM)\n");
  
  /* Infos sur les cellules du MNT */
  switch (mnt->cells) {
  case NIL_CELLS:
    printf ("\t(DTM cells not allocated...)\n");
    break;
  case CELLS_ALLOUEES:
    printf ("\t(DTM celles simply allocated)\n");
    break;
  case CELLS_DETERMINEES:
    printf ("\tSome cells have been merged (%d cells in total)\n",
	    mnt->nb_cells);
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }
  
  switch (mnt->luminance) {
  case NIL_LUMINANCES:
    printf ("\t(No luminance values are defined)\n");
    break;
  case LUMINANCES_DETERMINEES:
    printf ("\tThe luminance values are defined in the DTM\n");
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }

  /* Les bornes */
  switch (mnt->bornes) {
  case NIL_DTM_BORNES:
    printf ("\t(The bounds of the perceived area are not defined)\n");
    break;
  case BORNES_DTM_CALCULEES:
    printf ("\tBounds of the perceived area:\n");
    if (print_dtm_bornes (mnt) != DTM_OK)
      return (-1);
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }

  return (DTM_OK);
}

/**********************************************************************
 *
 * Affichage d'infos sur une position géodésique
 *
 **********************************************************************/

void
dtm_infoGeodesicPos (DTM_GEODESIC_POINT* pos)
{
  printf ("\t\tLat/Lon/Alt (deg/m) = %.5f %.5f %.2f\n",
	  RAD_TO_DEG(pos->lat),
	  RAD_TO_DEG(pos->lon),
	  pos->alt);
  printf ("\t\tEasting/Northing/Alt (m) = %.2f %.2f %.2f\n",
	  pos->easting,
	  pos->northing,
	  pos->alt);
  printf ("\t\t(nest East meridian (deg): %.2f)\n",
	  RAD_TO_DEG(pos->nextEastMeridian));
}
  
/**********************************************************************
 *
 * Affichage d'infos sur une cellule du MNT
 *
 **********************************************************************/

int
info_dtm_cell (DTM*		mnt,
	       int		i,
	       int		j)
     
{
  DTM_CELL*   cell;
  
  if ((i < 0) || (i >= mnt->nblig) || (j < 0) || (j >= mnt->nbcol)) {
    printf ("\n\t\tCellule (%d, %d) hors du modèle\n", i, j);
    printf ("\tCoordonnées de son centre / origine MNT : (%.3f, %.3f)\n",
	    (i + 0.5) * mnt->scale_x, (j + 0.5) * mnt->scale_y);
    printf ("\tCoordonnées de son centre / repère global : (%.3f, %.3f)\n",
	    (i + 0.5) * mnt->scale_x + mnt->x_origine,
	    (j + 0.5) * mnt->scale_y + mnt->y_origine);
    return (DTM_OK);
  }
  
  cell = mnt->cells_tab + (i * mnt->nbcol) + j;
  
  if (cell->state == EMPTY) 
    printf ("\n\t\tCellule (%d, %d) vide\n", i, j);
  else 
    printf ("\tCellule (%d, %d) remplie\n", i, j);

  /* Infos on the position of the cell wrt. the dtm and its frame */
  printf ("\tCoordonnées de son centre / origine MNT : (%.3f, %.3f)\n",
	  (i + 0.5) * mnt->scale_x, (j + 0.5) * mnt->scale_y);
  printf ("\tCoordonnées de son centre / repère global : (%.3f, %.3f)\n",
	  (i + 0.5) * mnt->scale_x + mnt->x_origine,
	  (j + 0.5) * mnt->scale_y + mnt->y_origine);
  
  /* Its geodesic position */
  if ((mnt->geoReference == GEO_REFERENCE_SET)
      && (mnt->geoOrientation != NO_GEO_ORIENTATION)) {
    DTM_GEODESIC_POINT pos;
    double xCell, yCell;
    pos.alt = cell->current_info.z_moyen;

    xCell = i * mnt->scale_x;
    yCell = j * mnt->scale_y;
    switch (mnt->geoOrientation) {
    case GEO_NORTH:
      pos.easting = - yCell + mnt->geodesicOrigin.easting;
      pos.northing = xCell + mnt->geodesicOrigin.northing;
      break;
    case GEO_WEST:
      pos.easting = - xCell + mnt->geodesicOrigin.easting;
      pos.northing = - yCell + mnt->geodesicOrigin.northing;
      break;
    case GEO_SOUTH:
      pos.easting = yCell + mnt->geodesicOrigin.easting;
      pos.northing = - xCell + mnt->geodesicOrigin.northing;
      break;
    case GEO_EAST:
      pos.easting = xCell + mnt->geodesicOrigin.easting;
      pos.northing = yCell + mnt->geodesicOrigin.northing;
      break;
    default:
      printf ("Nowhere !\n");
      return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
      /*     break; */
      /* Unreached */
    }
    pos.nextEastMeridian = mnt->geodesicOrigin.nextEastMeridian;
    dtm_wgs84utm2lla (pos.easting, pos.northing, pos.nextEastMeridian,
		      &(pos.lat), &(pos.lon));
    dtm_infoGeodesicPos (&pos);
  }

  /* Infos on what is in the cell */
  if (cell->state != EMPTY) {
    printf ("\tPremire date de mise  jour : %d\n", cell->current_info.firstDate);
    printf ("\tDernire date de mise  jour : %d\n", cell->current_info.lastDate);
    printf ("\t\tContient %d points\n", cell->current_info.nb_points);
    printf ("\t\tZ = %.3f, Sigma_Z = %.5f\n", cell->current_info.z_moyen, cell->current_info.sigma_z);
    printf ("\tLuminance moyenne : %d\n", cell->current_info.luminance);
    printf ("\tConfidence : %.3f\n", cell->current_info.confidence);
    printf ("\tPoids cumul des points : %.3f\n", cell->current_info.weight);
  }
  
  return (DTM_OK);
}


/**********************************************************************
 *
 *
 **********************************************************************/

int 
print_dtm_bornes (DTM*   mnt)
     
{
  /* !!! Introduire l'origine pr calculer le cadre perçu !!! */

  switch (mnt->stockage) {
  case X_LIGNES:
    printf ("\t\tIndices min et max de la zone perçue :\n");
    printf ("\t\tSuivant I (X) : de %d à %d (de %.3f à %.3f mètres)\n",
	    mnt->i_min, mnt->i_max,
	    (mnt->i_min + 0.5) * mnt->scale_x,
	    (mnt->i_max + 0.5) * mnt->scale_x);
    printf ("\t\tSuivant J (Y) : de %d à %d (de %.3f à %.3f mètres)\n",
	    mnt->j_min, mnt->j_max,
	    (mnt->j_min + 0.5) * mnt->scale_y,
	    (mnt->j_max + 0.5) * mnt->scale_y);
    break;
  case X_COLONNES:
    printf ("\t\tIndices min et max dela zone perçue :\n");
    printf ("\t\tSuivant I (Y) : de %d à %d (de %.3f à %.3f mètres)\n",
	    mnt->i_min, mnt->i_max,
	    (mnt->i_min + 0.5) * mnt->scale_y,
	    (mnt->i_max + 0.5) * mnt->scale_y);
    printf ("\t\tSuivant J (X) : de %d à %d (de %.3f à %.3f mètres)\n",
	    mnt->j_min, mnt->j_max,
	    (mnt->j_min + 0.5) * mnt->scale_x,
	    (mnt->j_max + 0.5) * mnt->scale_x);
    break;
  default:
    return dtmError (ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }

  printf ("\t\tAltitudes codées dans le mnt de %.3f à %.3f mètres\n",
	  mnt->z_min, mnt->z_max);
  printf ("\t\tNombres min et max de pts contenus dans les cellules:  (%d, %d)\n",
	  mnt->nbPtsMin, mnt->nbPtsMax);
  printf ("\t\tEcarts types codés dans le mnt de %f à %f mètres\n",
	  mnt->sigmaMin, mnt->sigmaMax);
  if (mnt->luminance == LUMINANCES_DETERMINEES) {
    printf ("\t\tCouleurs codées dans le mnt de %d à %d\n",
	    mnt->colorMin, mnt->colorMax);
  }
  
  return (DTM_OK);
}
