/*	$Id: quads.c,v 1.2 2006/12/01 16:05:23 simon Exp $ */

/* 
 * Copyright (c) 2002 LAAS/CNRS                       --  Thu Nov  2 2000
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "libdtm.h"

static void	scanEdge(int x0, int y0, double z0, double i0, int x1,
			int y1, double z1, double i1, DtmQuadSpan *span);
static void	midPoint(int x0, int y0, double z0, double i0, int x1,
			int y1, double z1, double i1, 
			void (*write)(int x,int y, double z, double i,
				      DtmQuadSpan *span),
			DtmQuadSpan *spans);
static void	spanwriteXY(int x, int y, double z, double i,
			DtmQuadSpan *span);
static void	spanwriteYX(int y, int x, double z, double i,
			DtmQuadSpan *span); 
static void	spanwriteXNY(int x, int y, double z, double i,
			DtmQuadSpan *span); 
static void	spanwriteNYX(int y, int x, double z, double i,
			DtmQuadSpan *span); 

/*
 * --- dtmInitQuads -----------------------------------------------------
 *
 * Initialize a quad structure to some reasonable default values.
 */

int
dtmInitQuads(DtmQuad *quad)
{
   quad->spans = 0;
   quad->span = NULL;
   quad->vertices = 0;
   quad->vd = NULL;
   quad->vx = quad->vy = NULL;
   quad->vz = quad->vi = NULL;
   return 0;
}


/*
 * --- dtmAllocQuads ----------------------------------------------------
 *
 * Allocate memory for internal arrays of the `quad' structure.
 * `width' should be equal to the length of dtm lines.
 */

int
dtmAllocQuads(DtmQuad *quad, int width)
{
   /* nobody's defined */
   quad->d[0] = quad->d[1] = quad->d[2] = quad->d[3] = 0;

   if (quad->vertices < width) {
      quad->vertices = 0;

      if (quad->vd) free(quad->vd);
      quad->vd = malloc(width * sizeof *quad->vd);
      if (quad->vx) free(quad->vx);
      quad->vx = malloc(width * sizeof *quad->vx);
      if (quad->vy) free(quad->vy);
      quad->vy = malloc(width * sizeof *quad->vy);
      if (quad->vz) free(quad->vz);
      quad->vz = malloc(width * sizeof *quad->vz);
      if (quad->vi) free(quad->vi);
      quad->vi = malloc(width * sizeof *quad->vi);
      
      if (!(quad->vd && quad->vx && quad->vy && quad->vz && quad->vi))
	 return 1;

      quad->vertices = width;
   }

   /* no vertex is buffered */
   memset(quad->vd, 0, width * sizeof *quad->vd);
   return 0;
}

void
dtmDeleteQuads(DtmQuad *quad)
{
   if (!quad) return;
   if (quad->vd) free(quad->vd);
   if (quad->vx) free(quad->vx);
   if (quad->vy) free(quad->vy);
   if (quad->vz) free(quad->vz);
   if (quad->vi) free(quad->vi);
}


/*
 * --- dtmMakeQuad ------------------------------------------------------
 *
 * Compute the list of cells intersected by the polygon currently stored
 * within the `quad' structure.
 * Returns total number of cells, compute number `ny' of lines and the
 * first line index `ymin'. Fills the `span' array of the quad structure
 * too.
 */

int
dtmMakeQuad(DtmQuad *quad, int *ymin, int *ny)
{
   int x0,y0, x1,y1;
   double z0,z1, i0,i1;
   int i, min,max, fvx;
   DtmQuadSpan *span;

   /* fvx = first defined vertex */
   for(fvx=0; !quad->d[fvx] && fvx<4; fvx++) /* empty body*/;
   if (fvx>=4) return 0;

   /* compute rectangular bounding box */
   min = max = quad->y[fvx];
   x0 = x1 = quad->x[fvx];
   for(i=fvx+1;i<4;i++) if (quad->d[i]) {
      if (quad->y[i]>max) max = quad->y[i];
      if (quad->y[i]<min) min = quad->y[i];
      if (quad->x[i]<x0) x0 = quad->x[i];
      if (quad->x[i]>x1) x1 = quad->x[i];
   }

   /* alloc more spans if needed */
   if (quad->spans < max-min+1) {
      if (quad->span) free(quad->span);
      quad->spans = max-min+1;
      quad->span = malloc(quad->spans * sizeof(DtmQuadSpan));
      if (!quad->span) { quad->spans = 0; return 0; }
   }

   /* handle degenerated quads (single cell) specially -- for efficiency */
   if (max-min+1 == 1 && x0 == x1) {
      quad->span[0].min = x0;
      quad->span[0].max = x1;
      quad->span[0].zmin = quad->z[fvx];
      quad->span[0].zmax = quad->z[fvx];
      quad->span[0].imin = quad->i[fvx];
      quad->span[0].imax = quad->i[fvx];

      *ymin = min;
      *ny = 1;
      return 1;
   }

   /* general case (height > 1 or width > 1) */

   /* reset all spans */
   for(i=0, span=quad->span;i<max-min+1;i++, span++) span->count = 0;

   /* loop on vertices */
   x0 = quad->x[fvx]; y0 = quad->y[fvx];
   z0 = quad->z[fvx]; i0 = quad->i[fvx];
   for(i=fvx+1; i<=fvx+4; i++) if (quad->d[i%4]) {
      x1 = quad->x[i%4]; y1 = quad->y[i%4];
      z1 = quad->z[i%4]; i1 = quad->i[i%4];

      scanEdge(x0,y0,z0,i0, x1,y1,z1,i1, quad->span - min);

      x0 = x1; y0 = y1;
      z0 = z1; i0 = i1;
   }

   *ymin = min;
   *ny = max-min+1;

   /* compute number of cells */
   for(i=max=0, span=quad->span;i<*ny;i++, span++)
      max += span->max - span->min + 1;

   return max;
}

static void
scanEdge(int x0, int y0, double z0, double i0,
	 int x1, int y1, double z1, double i1, DtmQuadSpan *span)
{
   int swp;
   double dswp;

   if (x1 < x0) { 
      swp = x1; x1 = x0; x0 = swp;
      swp = y1; y1 = y0; y0 = swp;
      dswp = z1; z1 = z0; z0 = dswp;
      dswp = i1; i1 = i0; i0 = dswp;
   }

   if (y1 >= y0) {
      if (y1-y0 > x1-x0)
	 midPoint(y0,x0,z0,i0, y1,x1,z1,i1, spanwriteYX, span);
      else
	 midPoint(x0,y0,z0,i0, x1,y1,z1,i1, spanwriteXY, span);
   } else {
      if (y0-y1 > x1-x0) 
	 midPoint(-y0,x0,z0,i0, -y1,x1,z1,i1, spanwriteNYX, span);
      else
	 midPoint(x0,-y0,z0,i0, x1,-y1,z1,i1, spanwriteXNY, span);
   }
}

static void
midPoint(int x0, int y0, double z0, double i0,
	 int x1, int y1, double z1, double i1,
	 void (*write)(int x,int y, double z, double i, DtmQuadSpan *span),
	 DtmQuadSpan *spans)
{
   int dx, dy;
   int d, incrE, incrNE;
   int x = x0;
   int y = y0;
   double z = z0, dz;
   double i = i0, di;

   write(x,y,z,i, spans);

   if (x1==x0) return;

   dx = x1 - x0;
   dy = y1 - y0;
   dz = (z1 - z0)/(x1 - x0);
   di = (i1 - i0)/(x1 - x0);
   d = 2*dy - dx;
   incrE = 2*dy;
   incrNE = 2*(dy-dx);

   while(x<x1) {
      if (d<=0) { 
	 d += incrE;
      } else {
	 d += incrNE;
	 y++;
      }
      x++; z+=dz; i+=di;
      write(x,y,z,i, spans);
   }
}

#define SPAN_WRITE(x,y,z,i, span)\
   if (span[y].count++) {	\
      if (span[y].min > x) {	\
	 span[y].min = x;	\
	 span[y].zmin = z;	\
	 span[y].imin = i;	\
      }				\
      if (span[y].max < x) {	\
	 span[y].max = x;	\
	 span[y].zmax = z;	\
	 span[y].imax = i;	\
      }				\
   } else {			\
      span[y].min = x;		\
      span[y].max = x;		\
      span[y].zmin = z;		\
      span[y].zmax = z;		\
      span[y].imin = i;		\
      span[y].imax = i;		\
   }

static void
spanwriteXY(int x, int y, double z, double i, DtmQuadSpan *span) {
   SPAN_WRITE(x,y,z,i,span);
}

static void
spanwriteYX(int y, int x, double z, double i, DtmQuadSpan *span) {
   SPAN_WRITE(x,y,z,i,span);
}

static void
spanwriteXNY(int x, int y, double z, double i, DtmQuadSpan *span) {
   SPAN_WRITE(x,-y,z,i,span);
}

static void
spanwriteNYX(int y, int x, double z, double i, DtmQuadSpan *span) {
   SPAN_WRITE(x,-y,z,i,span);
}

